/*
 KX High Capacity High Performance Epoll Socket Client Daemon Class
 ---------------------------------------------------------------------------
 $Id: $ $Revision: $
 $HeadURL: $
 Copyright(c) 1990 - 2018 KRKIM(DURUMUL) Allrights reserved.
 */

/**
 * @file
 * @brief 
 * @detail 
 */

/*
 * History:
 * ===========================================================================
 * 2018.09.30  KRKIM   Initialized.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/epoll.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <signal.h>

#include <list>
#include <queue>
#include <vector>
#include <pthread.h>

#include "network.h"
#include "epollcomn.h"
#ifdef USE_INTERDBPOOL
#include "dbpool.h"
#endif

#include "kxHashMap.h"
#include "kxThread.h"
#include "iocq.h"
#include "tokenizer.h"

#include "kxClientSession.h"
#include "kxPollService.h"
#include "kxClientService.h"
#include "kxtracelog.h"

#define EPOLL_NEW
//Ŭ̾Ʈ ̹Ƿ Ź ONESHOT ʴ´.
//#define EPOLL_NOT_USE_ONESHOT_CLIENT

using namespace std;

CKxClientService::CKxClientService() : CKxService()
{
    m_pPollService = NULL;
    m_queuemode = QUEUE_MODE_WORKQUEUE | QUEUE_MODE_USE_CLIENT;
    m_bAutoReconnect = 1;
    m_ReconnectTimeval.tv_sec = 0;//1970.01.01 κ  ð() TIMESTAMP ̴.
    m_ReconnectTimeval.tv_usec = 0;
    m_nReconnectInterval = 30;//ʴ: αװ ʹ ׿ ...
    m_nTryConnectCount = 0;
}

CKxClientService::~CKxClientService()
{
    Destroy();
}

void CKxClientService::SetAutoReconnect(int bONOFF)
{
    m_bAutoReconnect = bONOFF;
}

void CKxClientService::SetMaxClient(int maxclient)
{
    m_nMaxClient = maxclient;
    int system_limit_fdcount = get_limit_fd();
    kxTraceLog(0, "SetMaxClient: Request MaxClient(%d),system limit max fd(%d)\n", maxclient, system_limit_fdcount);
    if (m_nMaxClient > system_limit_fdcount){
        m_nMaxClient = system_limit_fdcount;
        kxTraceLog(0, "SetMaxClient: over limit max fd(%d)\n", system_limit_fdcount);
    }
}

void CKxClientService::SetPollService(CKxPollService *pPollService)
{
    m_pPollService = pPollService;
}

bool CKxClientService::Init(int maxclient, int maxthreadpool, int epolltimeout,int conntimeout)
{
    SetMaxClient(maxclient);
    SetMaxThreadPool(maxthreadpool);
    m_epolltimeout = epolltimeout;
    m_conntimeout = conntimeout;
    kxTraceLog(0, "[KxSockClientService] Init:maxclient=%d maxthreadpool=%d epolltimeout=%d conntimeout=%d \n",
        maxclient,maxthreadpool, m_epolltimeout,m_conntimeout );
    if (m_nMaxThreadPool > 0){
        InitThreadPool(maxthreadpool);
    }
    return true;
}

bool CKxClientService::InitSockEvent()
{
    int epollsize = m_nMaxClient, poolsize = m_nMaxClient;
    m_eventfd = init_epoll(epollsize);
    if (m_eventfd < 0)
    {
        kxTraceLog(0, "InitSockClient:init_epoll error\n");
        return false;
    }

    // init pool
    m_pEpollEvents = new struct epoll_event[sizeof (*m_pEpollEvents) * poolsize];
    if (m_pEpollEvents == NULL)
    {
        kxTraceLog(0, "InitSockClient:memory alloc failed.\n");
        return false;
    }
    return true;
}

void CKxClientService::FreeSockEvent()
{
    close(m_eventfd);// ۾ ٸ close() ȣؼ ݾ־ Ѵ. 
    if (m_pEpollEvents)
        delete[]m_pEpollEvents;
    m_pEpollEvents = NULL;
}

bool CKxClientService::Connect(CKxClientSession *currentuser)
{
    int rc;
    int clientfd;
	/* set to reconnect to host */
    if (currentuser->GetSockFD() == INVALID_SOCKET){
        currentuser->InitSocket();
    }
    setnonblock(clientfd,false);

    /* connect to host */
    rc = currentuser->Connect();
    if(rc < 0){
        //  #define ECONNREFUSED    111 /* Connection refused */  ٸ.
        //linux     : EIDRM = 43 ECONNREFUSED = 111 EPIPE = 32
        //ms-windows: EIDRM = 111,ECONNREFUSED = 107 
        //http://www.virtsync.com/c-error-codes-include-errno
		kxTraceLog(0,": host %s port %d rc = %d errno %d(%s)\n",
            currentuser->GetHost(),currentuser->GetPort(),rc,errno,strerror(errno));
        //ū ߸ fd 
        if (errno == EPIPE || errno == ECONNREFUSED || errno == 111){//EIDRM
            return false;
        }
        // ŷ̶ ٸ ӵǴ ¸ ǹ.
	}

    clientfd = currentuser->GetSockFD();
	setnonblock(clientfd,true);
	settcpnodelay(clientfd,true);

#ifndef EPOLL_NEW
    del_epollfd(m_eventfd, clientfd);
#else
    Epoll_Del(m_eventfd, clientfd);
#endif
    /*no one shot */
#ifndef EPOLL_NEW
    rc = add_epollfd(m_eventfd, clientfd,EPEVT_READ|EPEVT_HANGUP|EPEVT_ERROR|EPEVT_ET);
#else
    //rc = Epoll_Add(m_eventfd, clientfd, currentuser, EPEVT_READ | EPEVT_HANGUP | EPEVT_ERROR | EPEVT_ET);
#ifdef EPOLL_NOT_USE_ONESHOT_CLIENT
    rc = Epoll_Add(m_eventfd, clientfd, currentuser, EPEVT_READ | EPEVT_WRITE | EPEVT_ERROR | EPEVT_ET);
#else
    rc = Epoll_Add(m_eventfd, clientfd, currentuser, EPEVT_READ | EPEVT_WRITE | EPEVT_ERROR | EPEVT_ET | EPEVT_ONESHOT);
#endif

#endif
    if (rc < 0)
    {
        kxTraceLog(0, "Connect:Epoll_Add sockfd(%d) error errno %d(%s)\n",clientfd,errno,strerror(errno));
        return false;
    }

    kxTraceLog(0, "[KxClientService] Connect:Host=%s port=%d sockfd(%d)\n", currentuser->GetHost(),currentuser->GetPort(),clientfd);
    return true;
}

int CKxClientService::TryConnectSocketList()
{
    int conn = 0;
    int doit = 0;
    int retry = 0;
    //Ͽ ŸӾƿ ൵ (1ʳ) ´(ŷ)
    //׷ epollhup ̳ processpending ewouldblock ɸ.
    //׷ loop û⶧ ֱ翬 õð  ֱ Ѵ.
    //2018.10.05

    struct timeval now, start;
    gettimeofday(&now, NULL);

    if (m_ReconnectTimeval.tv_sec == 0){// ó  翬
        gettimeofday(&start, NULL);
        m_ReconnectTimeval = start;
        kxTraceLog(0, "ϲ /ڵ 翬 ɽ: timestamp = %ld,üũ=%d sec\n",
            m_ReconnectTimeval.tv_sec,m_nReconnectInterval);
    }
    else{
        if (now.tv_sec <= m_ReconnectTimeval.tv_sec + m_nReconnectInterval){
            return 0;
        }
    }

    CKxClientSession *currentuser = NULL;
    list < CKxSockSession* >::iterator iter;
    m_ClientSockList.Lock();
    for (iter = m_ClientSockList.GetClientSockListPtr()->begin();
        iter != m_ClientSockList.GetClientSockListPtr()->end(); ++iter)
    {
        currentuser = (CKxClientSession* )*iter;
        if(currentuser->IsConnected() == false){
            if (m_nTryConnectCount > 0){
                kxTraceLog(0, " 翬 õ: host %s port %d\n",
                    currentuser->GetHost(), currentuser->GetPort());
            }
            else if (m_nTryConnectCount == 0){
                kxTraceLog(0, "õ: host %s port %d\n",
                    currentuser->GetHost(), currentuser->GetPort());
            }
            retry++;
            if(Connect(currentuser)== true){
                conn++;
			}
		}
    }
    if (retry > 0){
        m_nTryConnectCount++;
        if (m_nTryConnectCount % 10 == 0){
            //10 * m_nReconnectInterval ʰ  , õ  üũð  ÷ش.
            //,ƿ   ӽõ ѹݺ α׸  ̹Ƿ 10̳ 1߰,
            // 10 ʹ 2а ,20кʹ 3а,30 ʹ 4а
            //⼭ 10 m_nReconnectInterval  60 츦 ϰ ð  ȴ.
            //׾  Ƴ(ֳ  ˽) ڵ Ŭ̾Ʈ ϵ ڵ ȴ.
            m_nReconnectInterval = m_nReconnectInterval * 2;
            if (m_nReconnectInterval > 1800)//0.5ð 
                m_nReconnectInterval = 1800;//0.5ð üũ
        }
    }

    gettimeofday(&now, NULL);
    m_ReconnectTimeval = now;
    m_ClientSockList.UnLock();
    return conn;
}

bool CKxClientService::DestroySockClient(CKxClientSession *currentuser)
{
    //Ŭ̾ƮϿ  ʴ´.
    //翬 ؼ.
    currentuser->CloseClient();
    return true;
}

//ξ  ȣ
//: 0  0 ƴϸ 
int CKxClientService::OnServiceBegin()
{
    
    return CKxService::OnServiceBegin();
}

//ξ  ȣ
int CKxClientService::OnServiceEnd()
{
    
    return CKxService::OnServiceEnd();
}

//ξ  ȣ
//: 0  0 ƴϸ 
int CKxClientService::OnThreadBegin(int nID)
{
    return CKxService::OnThreadBegin(nID);
}

//ξ  ȣ
int CKxClientService::OnThreadEnd(int nID)
{
    
    return CKxService::OnThreadEnd(nID);
}

int CKxClientService::OnServiceLoop(int threadid)
{
    int	i;
    int	clientfd;
    int rc,start;
    int	poolsize = m_nMaxClient;
    int	max_got_events;
    //int reconnected = 1;
    int connected = 0;
    CKxQueue *pQueue = NULL;

    start = OnServiceBegin();
    while (start == 0 && m_bServiceRunning)
    {
        if (m_bAutoReconnect || connected == 0){
            connected = TryConnectSocketList();
            connected = 1;
        }
        //翬 ÿ ŸӾƿ ش.
        max_got_events = epoll_wait(m_eventfd, m_pEpollEvents, poolsize, m_epolltimeout/*-1*/);
        for (i = 0; i < max_got_events; i++)
        {
            // ̺Ʈť ԷѴ.
            //int clientfd = ev.data.fd;
            CKxClientSession *currentuser = (CKxClientSession *)m_pEpollEvents[i].data.ptr;//m_ClientSockList.Find(clientfd);
            clientfd = currentuser->GetSockFD();
            if (currentuser == NULL)
                continue;
                //kxTraceLog(0, "[%d] currentuser->m_threadno=%d\n",
                //    i,currentuser->m_threadno);
            if (currentuser->m_threadno == -1){
                //char *remoteaddr = inet_ntoa(clientaddr.sin_addr);

                //   óϴ 带 ã´.					
                int thread_index = 0;
                thread_index = m_pThreadPool->FindMinClients();//m_ClientSockList.GetCount() % m_nMaxThreadPool;
                //kxTraceLog(0, "new client sock(%d) thread_index=%d\n",
                //    clientfd, thread_index);

                currentuser->m_threadno = thread_index;
                //currentuser->SetSockFD(clientfd,&clientaddr);
                //currentuser->InitContext(MAX_IOBUFFSIZE);
                //m_ClientSockList.Add(currentuser);
                //kxTraceLog(0, "new client sock(%d) GetThread\n", clientfd);

                CKxThreadCtx * lThread = m_pThreadPool->GetThread(thread_index);
                lThread->Lock();
                lThread->m_clientnum++;
                lThread->UnLock();
                //kxTraceLog(0, "new client sock(%d) currentuser=%x Thread[%d] clients=%d\n",
                //    clientfd, currentuser, thread_index, lThread->m_clientnum);

//#ifdef EPOLL_NEW
//                rc = Epoll_Add(m_eventfd, clientfd, currentuser, EPEVT_READ | EPEVT_WRITE | EPEVT_ET | EPEVT_ONESHOT);
//                if (rc < 0)
//                {
//                    kxTraceLog(0, "ERROR:add client to epool,so close client sock now.\n");
//                    currentuser->OnDisconnected();
//                    currentuser->CloseClient();
//                    break;
//                }
//#endif
                currentuser->OnConnected();
                kxTraceLog(0, "Connect OK:client sock(%d) currentuser=%x dispatch to Thread[%d] clients=%d\n",
                    clientfd, currentuser, thread_index, lThread->m_clientnum);
            }

            CKxThreadCtx *lThread = m_pThreadPool->GetThread(currentuser->m_threadno);
            kxTraceLog(0, "Client Epoll:sock(%d) dispatch to Thread[%d] clients=%d\n",
                clientfd, lThread->m_ord_id - 1, lThread->m_clientnum);

            KX_QUEUE_ITEM qitem;
            qitem.mask = KXQUEUE_MASK_USE_EVENTS;
            qitem.events.epollevent = m_pEpollEvents[i];

            //kxTraceLog(0, " PushSignal sock(%d) currentuser=%x\n", clientfd, currentuser);
            if (((m_queuemode & QUEUE_MODE_USE_CLIENT) & QUEUE_MODE_MASK) == QUEUE_MODE_USE_CLIENT){
                pQueue = (CKxQueue *)lThread;
            }
            else if (((m_queuemode & QUEUE_MODE_USE_SERVER) & QUEUE_MODE_MASK) == QUEUE_MODE_USE_SERVER){
                pQueue = this->GetQueue();
            }
            pQueue->PushSignal(qitem);
        }
        usleep(1);//FOR BIT CPU IDLE

        //TODO::Send pending data
    }
    kxTraceLog(0, "ServiceRunning Stopped.\n");
    rc = OnServiceEnd();
    return true;
}

int CKxClientService::OnThreadLoop(int threadid)
{
    int rc;
    int index,start;
    int count, count1;

    CKxThreadCtx * lThread;
    CKxService *pService;
    KX_QUEUE_ITEM qitem; //struct epoll_event ev;
    CKxQueue *pQueue = NULL;

    if (m_pThreadPool)
        lThread = m_pThreadPool->GetThreadByID(threadid);

    start = OnThreadBegin(lThread->m_ord_id);

    pService = (CKxService *)lThread->m_pService;//same as this
    kxTraceLog(TL_LOW, "[Thread %d] pthread_self() = %x thr_id = %x (%x) started\n",
        lThread->m_ord_id, pthread_self(), lThread->m_thr_id,threadid);

    //lThread = (CKxThreadCtx *)arg;
    //lThread->WatchJob();
    /*pthread_cond_wait() Ǻ cond ñ׳ ޵Ǳ⸦ ٸ.
       Լ ȣǸ ڵ mutex ǵش(UnLock m_mut internally)
       pthread_unlock_mutex() ȣѴ.
       Լ ȣ  Ǻ ñ׳ ޵  ٸ,
      ̶ CPU Һ ʴ´.
      ñ׳ ޹޾Ƽ 尡  ڵ mutex ´(
      pthread_lock_mutex() )
      */
    //TODO::Check This!!
    //Ǻ m_cond 羲  ϰ Ǹ 
    // ϳ  ̺Ʈ   óϴ° ƴѰ?
    //׷ٸ Ǯ   ְ   常  ۾ϴ° ƴѰ??
    //,庰 Ǻ ΰ ؽ óϿ,̺Ʈť ,
    //浹 ʵϸ ϰ Ѵ?
    //׷  ۵ 尡 ŵŸ óð  ,߿  
    //óð  ª Ŭ̾Ʈ ϴ Ÿ  ٲ ɼ ִ° .
    //̷ Ǹ 尣 ȭ  ó  Ͱ.
    //,  óϴ Ͱ  츸 ó ϰ ̺Ʈť  ܵó  ʰ?
    //http://www.joinc.co.kr/modules/moniwiki/wiki.php/Site/Thread ؼ м  Ұ

    while (start == 0 && m_bServiceRunning)
    {
        //Pop Queue Data
        //Subscribe mode or workqueue mode
        //fanout ( 庰 ü queue ó-)
        //  =>  尡  ڽ ť   ͸ ͼ  ͸ йó
        //work queue ( 庰  queue-ٸ ó)
        //  =>  尡  δٸ ť ͼ ó
        
        //Subscribe mode or workqueue mode
        //workqueue:  1  δٸ ϵ ť ͼ ó
        if (((m_queuemode & QUEUE_MODE_USE_CLIENT) & QUEUE_MODE_MASK) == QUEUE_MODE_USE_CLIENT){
            pQueue = (CKxQueue *)lThread;
        }
        else if (((m_queuemode & QUEUE_MODE_USE_SERVER) & QUEUE_MODE_MASK) == QUEUE_MODE_USE_SERVER){
            pQueue = pService->GetQueue();
        }
        kxTraceLog(TL_LOW, "[Thread %d] WaitCondSignal() threadid=%x pQueue=%x(mode=%x)\n",lThread->m_ord_id,threadid,pQueue,m_queuemode);
        pQueue->WaitCondSignal();
        while (1){
            pQueue->Lock();
            count = pQueue->GetQueueSize();
            if (pQueue->IsEmpty() || count < 1){
                //kxTraceLog(TL_LOW,"[CKxThreadCtx %d] queue count=%d suspended to wait signal.\n",m_ord_id,count);
                pQueue->UnLock();
                break;
            }
			
			qitem = pQueue->PopData(0);
			count1 = pQueue->GetQueueSize();
			pQueue->UnLock();
            //kxTraceLog(TL_LOW, "[Thread %d] thr_id = %x popup ev,pService=%x\n", lThread->m_ord_id, lThread->m_thr_id, pService);
            pService->OnThreadProc(lThread->m_ord_id,(void*)&qitem);
            pQueue->Free(&qitem);//2018.09.30
        }
    }
    kxTraceLog(TL_LOW, "[Thread %d] pthread_self() = %x thr_id = %x stop\n", lThread->m_ord_id, pthread_self(), lThread->m_thr_id);
    rc = OnThreadEnd(lThread->m_ord_id);
    return 0;
}

bool CKxClientService::StartService()
{
    int connected;
    InitSockEvent();

    return CKxService::StartService();
}

bool CKxClientService::StopService()
{
    SetServiceRun(false);
    FreeSockEvent();
    return CKxService::StopService();
}

// Ŭ̾Ʈ , Ŭ̾Ʈ Ѵ.
CKxClientSession * CKxClientService::AddClientSock(char *hostname,int port,int conntimeout)
{
    bool rc = false;
    CKxClientSession *currentuser = new CKxClientSession();
   
    currentuser->Init(hostname, port, conntimeout);
    if (currentuser != NULL)
    {
        currentuser->InitContext(MAX_IOBUFFSIZE);
        m_ClientSockList.Add(currentuser);
        return currentuser;
    }
    return NULL;
}

//̹   accept 
//CKxClientSession * CKxClientService::AddClientSock(int sockfd, sockaddr_in *pClientAddr)
//{
//    bool rc = false;
//    CKxClientSession *currentuser = new CKxClientSession();
//    if (currentuser != NULL)
//    {
//        currentuser->SetSockFD(sockfd, pClientAddr);
//        currentuser->InitContext(MAX_IOBUFFSIZE);
//        m_ClientSockList.Add(currentuser);
//        return currentuser;
//    }
//    return NULL;
//}

bool CKxClientService::RemoveClientSock(int sockfd)
{
    return m_ClientSockList.Remove(sockfd);
}

bool CKxClientService::RemoveClientSock(CKxClientSession *pUser)
{
    return m_ClientSockList.Remove(pUser);
}

int CKxClientService::OnServicePolling(TIPARAM *polltime)
{
	return CKxService::OnServicePolling(polltime);
}

int CKxClientService::OnThreadQueue(int nID/*orderid*/,void *param/*popped queue data*/)
{
    return CKxService::OnThreadQueue(nID,param);
}

//http://vpj.posterous.com/tcp-echo-server-example-in-c-using-epoll
//int CKxClientService::OnThreadProc(struct epoll_event ev)
int CKxClientService::OnThreadProc(int nID/*orderid*/,void *param/*popped queue data*/)
{
    // read from the cfd
    int rc = 0;
    int rc2 = 0;
    KX_QUEUE_ITEM *qitem = (KX_QUEUE_ITEM*) param;
    struct epoll_event ev;
    ev = qitem->events.epollevent;
    int clientfd;
    CKxClientSession *currentuser = (CKxClientSession *)ev.data.ptr;//m_ClientSockList.Find(clientfd);
    CKxThreadCtx * lThread;
    CKxService *pService;

    /*lThread = m_pThreadPool->GetThreadByOrderID(nID);*/
    lThread = m_pThreadPool->GetThread(currentuser->m_threadno);
    pService = (CKxService *)lThread->m_pService;//same as this

    clientfd = currentuser->GetSockFD();
    //kxTraceLog(0, "[KxClientService] OnThreadProc:client sock(%d) currentuser=%x ev.data.fd=%x\n", clientfd, currentuser,ev.data.fd);

    if (currentuser == NULL)
    {
        kxTraceLog(0, "error,client sock_context is null\n");
        return -1000;
    }

    if (ev.events & EPOLLHUP){
        CloseClient(currentuser);

        kxTraceLog(0, "client(%d) EPOLLHUP detected.close socket.Thread[%d] clients= %d\n",
            clientfd, currentuser->m_threadno, lThread->m_clientnum);
        return -1002;
    }

    if (ev.events & EPOLLERR){
        kxTraceLog(0, "client(%d) EPOLLERR detected.so skip.\n", clientfd);
        return -1003;
    }

    if (ev.events & EPOLLIN){
        kxTraceLog(0, "client(%d) EPOLLIN status detected.available to recv incoming data now.\n", clientfd);
        currentuser->Lock();
        rc = currentuser->OnRecvPacket();
        currentuser->UnLock();
        if (rc == SOCK_TRY_CLOSE){
            CloseClient(currentuser);
            //int trid = 123;
            //int duringsec = 60;
            //char pkstr[1031];
            //system hangup
            //sprintf(pkstr, "HUP %d %d\n", cfd,trid,duringsec);
            //senddatatoall(pkstr);
            kxTraceLog(0, "client(%d) SOCK_TRY_CLOSE\n",clientfd);
            return -1004;
        }
        // re add to epoll on ET mode (in oneshot)
#ifndef EPOLL_NOT_USE_ONESHOT_CLIENT
#ifndef EPOLL_NEW
        rc2 = modify_epollfd(m_eventfd,clientfd,EPEVT_READ|EPEVT_WRITE|EPEVT_ET|EPEVT_ONESHOT);
#else
        rc2 = Epoll_Modify(m_eventfd, clientfd, currentuser, EPEVT_READ | EPEVT_WRITE | EPEVT_ET | EPEVT_ONESHOT);
#endif
#endif

        if (rc2 < 0)
        {
            kxTraceLog(0, "client(%d) ERROR:modify_epollfd to recv.\n", clientfd);
        }
    }
    if (ev.events & EPOLLOUT){
        kxTraceLog(0, "client(%d) EPOLLOUT status detected.available to send now.\n", clientfd);
        currentuser->Lock();
        rc = currentuser->OnSendPacket();
        currentuser->UnLock();
        if (rc == SOCK_TRY_CLOSE){
            CloseClient(currentuser);
            kxTraceLog(0, "client(%d) EPOLLOUT finish DATA,close socket.Thread[%d] clients= %d\n",
                clientfd, currentuser->m_threadno, lThread->m_clientnum);
        }
        else if (rc == SOCK_DATA_PENDING_SHIFT_SENDBUFFER){
            kxTraceLog(0, "client(%d) send buffer is remain,so modify_epollfd to send at nexttime.\n", clientfd);
#ifndef EPOLL_NOT_USE_ONESHOT_CLIENT
#ifndef EPOLL_NEW
            rc2 = modify_epollfd(m_eventfd,clientfd,EPEVT_READ|EPEVT_WRITE|EPEVT_ET|EPEVT_ONESHOT);
#else			
            rc2 = Epoll_Modify(m_eventfd, clientfd, currentuser, EPEVT_READ | EPEVT_WRITE | EPEVT_ET | EPEVT_ONESHOT);
#endif
#endif
            if (rc2 < 0)
            {
                kxTraceLog(0, "client(%d) ERROR:modify_epollfd to send.\n", clientfd);
            }
        }
        else{//to read event only
            kxTraceLog(0, "client(%d) send buffer is flush,so modify_epollfd to recv at nexttime.\n", clientfd);
#ifndef EPOLL_NOT_USE_ONESHOT_CLIENT
#ifndef EPOLL_NEW
            rc2 = modify_epollfd(m_eventfd,clientfd,EPEVT_READ|EPEVT_ET|EPEVT_ONESHOT);
#else
            rc2 = Epoll_Modify(m_eventfd, clientfd, currentuser, EPEVT_READ | EPEVT_ET | EPEVT_ONESHOT);
            //rc2 = Epoll_Modify(m_eventfd, clientfd, currentuser, EPEVT_READ | EPEVT_WRITE | EPEVT_ET | EPEVT_ONESHOT);
#endif
#endif
            if (rc2 < 0)
            {
                kxTraceLog(0, "client(%d) ERROR:modify_epollfd to recv after send.\n", clientfd);
            }
        }
    }
    return rc;
}

bool CKxClientService::senddatatoall(const char *chatmsg, int len)
{
    int sendflags = 0;
    CKxClientSession *currentuser = NULL;
    list < CKxSockSession* >::iterator iter;
    m_ClientSockList.Lock();
    for (iter = m_ClientSockList.GetClientSockListPtr()->begin();
        iter != m_ClientSockList.GetClientSockListPtr()->end(); ++iter)
    {
        currentuser = (CKxClientSession* )*iter;
        // sending only / no need to lock
        send((*iter)->GetSockFD(), chatmsg, len, sendflags);
    }
    m_ClientSockList.UnLock();
    return true;
}

int CKxClientService::PostSendToAll(char *pData, int nLen,int bNow)
{
    int sendflags = 0;
    CKxClientSession *currentuser = NULL;
    list < CKxSockSession* >::iterator iter;
    m_ClientSockList.Lock();
    for (iter = m_ClientSockList.GetClientSockListPtr()->begin();
        iter != m_ClientSockList.GetClientSockListPtr()->end(); ++iter)
    {
        currentuser = (CKxClientSession* )*iter;
        currentuser->PostSend(pData, nLen);
        if (bNow)
            currentuser->SendPacket();
        //kxTraceLog(0, "PostSendToAll\n");
    }
    m_ClientSockList.UnLock();
    return true;
}

int CKxClientService::PostSendToDataChannel(char *pData, int nLen,int bNow)
{
    return 0;
}

/**
 *
 * @brief   ش Ŭ̾Ʈ ݴ´.
 * @details
 * @param   CKxClientSession pclient Ŭ̾Ʈ  Ŭ
 * @date    2018-09-09
 * @version 0.0.1
 * @return  0 -> , -1 -> 
 * @todo
 */
int CKxClientService::CloseClient(CKxClientSession *pclient)
{
    int clientfd;
    clientfd = pclient->GetSockFD();
    pclient->OnDisconnected();
#ifndef EPOLL_NEW
    del_epollfd(m_eventfd, clientfd);
#else
    Epoll_Del(m_eventfd, clientfd);
    CKxThreadCtx * lThread = m_pThreadPool->GetThread(pclient->m_threadno);
    lThread->Lock();
    lThread->m_clientnum--;
    lThread->UnLock();
#endif
	kxTraceLog(0,"[CloseClient] ϲ: host %s port %d\n",pclient->GetHost(),pclient->GetPort());

    if (m_bAutoReconnect){
        pclient->CloseClient();
    }
    else{
        RemoveClientSock(clientfd);
        pclient->CloseClient();
        delete pclient;
    }
    return 0;
}
